#!/usr/bin/env python3
"""
Windows Disk Space Analyzer (System Management Tool)

Parameters:
    --analyze PATH: Analyze disk space usage for path
    --top N: Show top N largest files/folders
    --type EXT: Filter by file extension
    --min-size SIZE: Filter by minimum file size
    --export FILE: Export results to file
    --format FORMAT: Output format (table/json)

Requires: psutil, tabulate packages
Warning: Analysis of large directories may take significant time
"""

import os
import sys
import hashlib
import json
import argparse
import time
from typing import List, Dict, Any, Optional, Tuple
from tabulate import tabulate
import logging
from pathlib import Path
import concurrent.futures
import shutil

class DiskSpaceAnalyzer:
    def __init__(self, max_threads: int = 10):
        self.max_threads = max_threads
        self.logger = logging.getLogger(__name__)
        self.duplicates: Dict[str, List[str]] = {}
        self.large_files: List[Tuple[str, int]] = []
        self.dir_sizes: Dict[str, int] = {}

    def get_file_hash(self, filepath: str, chunk_size: int = 8192) -> str:
        """Calculate SHA-256 hash of a file."""
        sha256_hash = hashlib.sha256()
        try:
            with open(filepath, "rb") as f:
                for byte_block in iter(lambda: f.read(chunk_size), b""):
                    sha256_hash.update(byte_block)
            return sha256_hash.hexdigest()
        except Exception as e:
            self.logger.error(f"Error hashing file {filepath}: {str(e)}")
            return ""

    def analyze_directory(self, path: str, min_size: int = 100*1024*1024,
                        find_duplicates: bool = True) -> Dict[str, Any]:
        """Analyze disk space usage in a directory."""
        result = {
            'total_size': 0,
            'file_count': 0,
            'dir_count': 0,
            'large_files': [],
            'duplicates': {},
            'dir_sizes': {}
        }
        
        try:
            # Use ThreadPoolExecutor for parallel processing
            with concurrent.futures.ThreadPoolExecutor(max_workers=self.max_threads) as executor:
                futures = []
                
                for root, dirs, files in os.walk(path):
                    result['dir_count'] += len(dirs)
                    result['file_count'] += len(files)
                    
                    dir_size = 0
                    for file in files:
                        filepath = os.path.join(root, file)
                        try:
                            size = os.path.getsize(filepath)
                            dir_size += size
                            
                            # Track large files
                            if size >= min_size:
                                result['large_files'].append((filepath, size))
                            
                            # Calculate file hashes for duplicate detection
                            if find_duplicates and size > 0:
                                futures.append(
                                    executor.submit(self.get_file_hash, filepath)
                                )
                                
                        except Exception as e:
                            self.logger.error(
                                f"Error processing file {filepath}: {str(e)}"
                            )
                    
                    result['dir_sizes'][root] = dir_size
                    result['total_size'] += dir_size
                
                # Process file hashes and find duplicates
                if find_duplicates:
                    hash_to_files = {}
                    for future in concurrent.futures.as_completed(futures):
                        try:
                            file_hash = future.result()
                            if file_hash:
                                if file_hash in hash_to_files:
                                    if file_hash not in result['duplicates']:
                                        result['duplicates'][file_hash] = [
                                            hash_to_files[file_hash]
                                        ]
                                    result['duplicates'][file_hash].append(
                                        filepath
                                    )
                                else:
                                    hash_to_files[file_hash] = filepath
                        except Exception as e:
                            self.logger.error(
                                f"Error processing file hash: {str(e)}"
                            )
                            
        except Exception as e:
            self.logger.error(f"Error analyzing directory {path}: {str(e)}")
            
        return result

    def format_size(self, size: int) -> str:
        """Format size in bytes to human-readable format."""
        for unit in ['B', 'KB', 'MB', 'GB', 'TB']:
            if size < 1024.0:
                return f"{size:.2f} {unit}"
            size /= 1024.0
        return f"{size:.2f} PB"

    def display_results(self, results: Dict[str, Any],
                       output_format: str = 'table',
                       top_n: int = 10):
        """Display analysis results in the specified format."""
        if output_format == 'json':
            # Convert results to JSON-friendly format
            json_results = {
                'total_size': self.format_size(results['total_size']),
                'file_count': results['file_count'],
                'dir_count': results['dir_count'],
                'large_files': [
                    {'path': path, 'size': self.format_size(size)}
                    for path, size in sorted(
                        results['large_files'],
                        key=lambda x: x[1],
                        reverse=True
                    )[:top_n]
                ],
                'top_directories': [
                    {'path': path, 'size': self.format_size(size)}
                    for path, size in sorted(
                        results['dir_sizes'].items(),
                        key=lambda x: x[1],
                        reverse=True
                    )[:top_n]
                ],
                'duplicates': {
                    h: files for h, files in results['duplicates'].items()
                }
            }
            print(json.dumps(json_results, indent=4))
            
        else:
            # Display summary
            print("\n=== Disk Space Analysis Summary ===")
            print(f"Total Size: {self.format_size(results['total_size'])}")
            print(f"Total Files: {results['file_count']}")
            print(f"Total Directories: {results['dir_count']}")
            
            # Display largest files
            if results['large_files']:
                print("\n=== Largest Files ===")
                large_files_data = [
                    [path, self.format_size(size)]
                    for path, size in sorted(
                        results['large_files'],
                        key=lambda x: x[1],
                        reverse=True
                    )[:top_n]
                ]
                print(tabulate(
                    large_files_data,
                    headers=['Path', 'Size'],
                    tablefmt='grid'
                ))
            
            # Display largest directories
            print("\n=== Largest Directories ===")
            dir_data = [
                [path, self.format_size(size)]
                for path, size in sorted(
                    results['dir_sizes'].items(),
                    key=lambda x: x[1],
                    reverse=True
                )[:top_n]
            ]
            print(tabulate(
                dir_data,
                headers=['Path', 'Size'],
                tablefmt='grid'
            ))
            
            # Display duplicate files
            if results['duplicates']:
                print("\n=== Duplicate Files ===")
                for file_hash, file_list in results['duplicates'].items():
                    if len(file_list) > 1:
                        print(f"\nDuplicate set (hash: {file_hash[:8]}...):")
                        for filepath in file_list:
                            print(f"  {filepath}")

def main():
    parser = argparse.ArgumentParser(
        description='Disk Space Analyzer - Analyze disk space usage')
    parser.add_argument('path',
                       help='Directory path to analyze')
    parser.add_argument('--min-size', type=int, default=100*1024*1024,
                       help='Minimum size (bytes) for large files (default: 100MB)')
    parser.add_argument('--no-duplicates', action='store_true',
                       help='Skip duplicate file detection')
    parser.add_argument('--threads', type=int, default=10,
                       help='Maximum number of threads (default: 10)')
    parser.add_argument('--format', choices=['table', 'json'],
                       default='table',
                       help='Output format')
    parser.add_argument('--top', type=int, default=10,
                       help='Number of top items to display (default: 10)')
    
    args = parser.parse_args()
    
    # Setup logging
    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s - %(levelname)s - %(message)s'
    )
    
    try:
        # Validate path
        if not os.path.exists(args.path):
            print(f"Error: Path '{args.path}' does not exist")
            return 1
            
        analyzer = DiskSpaceAnalyzer(max_threads=args.threads)
        
        print(f"Analyzing directory: {args.path}")
        print("This may take a while depending on the directory size...")
        start_time = time.time()
        
        results = analyzer.analyze_directory(
            args.path,
            args.min_size,
            not args.no_duplicates
        )
        
        analysis_time = time.time() - start_time
        print(f"\nAnalysis completed in {analysis_time:.2f} seconds")
        
        analyzer.display_results(results, args.format, args.top)
        
    except Exception as e:
        logging.error(f"An error occurred: {str(e)}")
        return 1
        
    return 0

if __name__ == "__main__":
    exit(main())
